@@ -1,3 +1,35 @@
+/*
+   Copyright (C) 2002-2009 Karl J. Runge <runge@karlrunge.com> 
+   All rights reserved.
+
+This file is part of x11vnc.
+
+x11vnc is free software; you can redistribute it and/or modify
+it under the terms of the GNU General Public License as published by
+the Free Software Foundation; either version 2 of the License, or (at
+your option) any later version.
+
+x11vnc is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+GNU General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with x11vnc; if not, write to the Free Software
+Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA
+or see <http://www.gnu.org/licenses/>.
+
+In addition, as a special exception, Karl J. Runge
+gives permission to link the code of its release of x11vnc with the
+OpenSSL project's "OpenSSL" library (or with modified versions of it
+that use the same license as the "OpenSSL" library), and distribute
+the linked executables.  You must obey the GNU General Public License
+in all respects for all of the code used other than "OpenSSL".  If you
+modify this file, you may extend this exception to your version of the
+file, but you are not obligated to do so.  If you do not wish to do
+so, delete this exception statement from your version.
+*/
+
 /* -- xevents.c -- */
 
 #include "x11vnc.h"
@@ -44,7 +76,8 @@ void set_server_input(rfbClientPtr cl, int s);
 void set_text_chat(rfbClientPtr cl, int l, char *t);
 int get_keyboard_led_state_hook(rfbScreenInfoPtr s);
 int get_file_transfer_permitted(rfbClientPtr cl);
-void get_prop(char *str, int len, Atom prop);
+void get_prop(char *str, int len, Atom prop, Window w);
+int guess_dm_gone(int t1, int t2);
 
 static void initialize_xevents(int reset);
 static void print_xevent_bases(void);
@@ -54,18 +87,28 @@ static void grab_buster_watch(int parent, char *dstr);
 
 
 void initialize_vnc_connect_prop(void) {
+	char *prop_str;
 	vnc_connect_str[0] = '\0';
 	RAWFB_RET_VOID
 #if !NO_X11
+	prop_str = getenv("VNC_CONNECT");
+	if (prop_str == NULL) {
+		prop_str = "VNC_CONNECT";
+	}
 	vnc_connect_prop = XInternAtom(dpy, "VNC_CONNECT", False);
 #endif
 }
 
 void initialize_x11vnc_remote_prop(void) {
+	char *prop_str;
 	x11vnc_remote_str[0] = '\0';
 	RAWFB_RET_VOID
 #if !NO_X11
-	x11vnc_remote_prop = XInternAtom(dpy, "X11VNC_REMOTE", False);
+	prop_str = getenv("X11VNC_REMOTE");
+	if (prop_str == NULL) {
+		prop_str = "X11VNC_REMOTE";
+	}
+	x11vnc_remote_prop = XInternAtom(dpy, prop_str, False);
 #endif
 }
 
@@ -87,6 +130,189 @@ void initialize_clipboard_atom(void) {
 #endif	/* NO_X11 */
 }
 
+/*
+      we observed these strings:
+
+      6 gdm_string: Gnome-power-manager
+      6 gdm_string: Gnome-session
+      6 gdm_string: Gnome-settings-daemon
+      6 gdm_string: Login Window
+      6 gdm_string: Notify-osd
+      6 gdm_string: Panel
+     12 gdm_string: Metacity
+     12 gdm_string: gnome-power-manager
+     12 gdm_string: gnome-session
+     12 gdm_string: gnome-settings-daemon
+     12 gdm_string: notify-osd
+     18 gdm_string: Gdm-simple-greeter
+     24 gdm_string: metacity
+     36 gdm_string: gdm-simple-greeter
+ */
+
+static int gdm_string(char *str) {
+	if (str == NULL) {
+		return 0;
+	}
+	if (str[0] == '\0') {
+		return 0;
+	}
+	if (0) fprintf(stderr, "gdm_string: %s\n", str);
+	if (strstr(str, "gdm-") == str || strstr(str, "Gdm-") == str) {
+		if (strstr(str, "-greeter") != NULL) {
+			return 1;
+		}
+	}
+	return 0;
+}
+
+static int gdm_still_running(void) {
+#if NO_X11
+	return 0;
+#else
+	Window r, parent;
+	Window *winlist;
+	unsigned int nc;
+	int rc, i;
+	static XClassHint *classhint = NULL;
+	XErrorHandler old_handler;
+	int saw_gdm_name = 0;
+
+	/* some times a window can go away before we get to it */
+	trapped_xerror = 0;
+	old_handler = XSetErrorHandler(trap_xerror);
+
+	if (! classhint) {
+		classhint = XAllocClassHint();
+	}
+
+	/* we are xlocked. */
+	rc = XQueryTree_wr(dpy, DefaultRootWindow(dpy), &r, &parent, &winlist, &nc);
+	if (!rc || winlist == NULL || nc == 0) {
+		nc = 0;
+	}
+	for (i=0; i < (int) nc; i++) {
+		char *name = NULL;
+		Window w = winlist[i];
+		if (XFetchName(dpy, w, &name) && name != NULL) {
+			saw_gdm_name += gdm_string(name);
+			XFree_wr(name);
+		}
+		classhint->res_name = NULL;
+		classhint->res_class = NULL;
+		if (XGetClassHint(dpy, w, classhint)) {
+			name = classhint->res_name;
+			if (name != NULL) {
+				saw_gdm_name += gdm_string(name);
+				XFree_wr(name);
+			}
+			name = classhint->res_class;
+			if (name != NULL) {
+				saw_gdm_name += gdm_string(name);
+				XFree_wr(name);
+			}
+		}
+		if (saw_gdm_name > 0) {
+			break;
+		}
+	}
+	if (winlist != NULL) {
+		XFree_wr(winlist);
+	}
+
+	XSync(dpy, False);
+	XSetErrorHandler(old_handler);
+	trapped_xerror = 0;
+
+	return saw_gdm_name;
+#endif
+}
+
+static int wm_running(void) {
+	char *s = getenv("DEBUG_WM_RUNNING");
+	RAWFB_RET(0)
+#if NO_X11
+	return 0;
+#else
+	/*
+	 * Unfortunately with recent GDM (v2.28), they run gnome-session,
+	 * dbus-launch, and metacity for the Login greeter!  So the simple
+	 * XInternAtom checks below no longer work.
+	 */
+	if (gdm_still_running()) {
+		return 0;
+	}
+
+	/* we are xlocked. */
+	if (XInternAtom(dpy, "_NET_SUPPORTED", True) != None) {
+		if (s) rfbLog("wm is running (_NET_SUPPORTED).\n");
+		return 1;
+	}
+	if (XInternAtom(dpy, "_WIN_PROTOCOLS", True) != None) {
+		if (s) rfbLog("wm is running (_WIN_PROTOCOLS).\n");
+		return 1;
+	}
+	if (XInternAtom(dpy, "_XROOTPMAP_ID", True) != None) {
+		if (s) rfbLog("wm is running (_XROOTPMAP_ID).\n");
+		return 1;
+	}
+	if (XInternAtom(dpy, "_MIT_PRIORITY_COLORS", True) != None) {
+		if (s) rfbLog("wm is running (_MIT_PRIORITY_COLORS).\n");
+		return 1;
+	}
+	if (s) rfbLog("wm is not running.\n");
+	return 0;
+#endif	/* NO_X11 */
+	
+}
+
+int guess_dm_gone(int t1, int t2) {
+	int wait = t2;
+	char *avoid = getenv("X11VNC_AVOID_WINDOWS");
+	time_t tcheck = last_client;
+
+	if (last_open_xdisplay > last_client) {
+		/* better time for -display WAIT:... */
+		tcheck = last_open_xdisplay;
+	}
+
+	if (avoid && !strcmp(avoid, "never")) {
+		return 1;
+	}
+	if (!screen || !screen->clientHead) {
+		return 0;
+	}
+	if (avoid) {
+		int n = atoi(avoid);
+		if (n > 1) {
+			wait = n;
+		} else {
+			wait = 90;
+		}
+	} else {
+		static time_t saw_wm = 0;
+
+		wait = t2;
+
+		X_LOCK;
+		if (wm_running()) {
+			if (saw_wm == 0) {
+				saw_wm = time(NULL);
+			} else if (time(NULL) <= saw_wm + 2) {
+				/* try to wait a few seconds after transition */
+				;
+			} else {
+				wait = t1;
+			}
+		}
+		X_UNLOCK;
+	}
+	/* we assume they've logged in OK after wait seconds... */
+	if (time(NULL) <= tcheck + wait)  {
+		return 0;
+	}
+	return 1;
+}
+
 static void initialize_xevents(int reset) {
 #if NO_X11
 	RAWFB_RET_VOID
@@ -124,17 +350,27 @@ static void initialize_xevents(int reset) {
 		X_LOCK;
 		xselectinput_rootwin |= PropertyChangeMask;
 		XSelectInput_wr(dpy, rootwin, xselectinput_rootwin);
+
+		if (subwin && freeze_when_obscured) {
+			XSelectInput_wr(dpy, subwin, VisibilityChangeMask);
+		}
 		X_UNLOCK;
 		did_xselect_input = 1;
 	}
 	if (watch_selection && !did_xcreate_simple_window) {
 		/* create fake window for our selection ownership, etc */
 
-		X_LOCK;
-		selwin = XCreateSimpleWindow(dpy, rootwin, 0, 0, 1, 1, 0, 0, 0);
-		X_UNLOCK;
-		did_xcreate_simple_window = 1;
-		if (0) rfbLog("selwin: 0x%lx\n", selwin);
+		/*
+		 * We try to delay creating selwin until we are past
+		 * any GDM, (or other KillInitClients=true) manager.
+		 */
+		if (guess_dm_gone(5, 45)) {
+			X_LOCK;
+			selwin = XCreateSimpleWindow(dpy, rootwin, 3, 2, 1, 1, 0, 0, 0);
+			X_UNLOCK;
+			did_xcreate_simple_window = 1;
+			if (! quiet) rfbLog("created selwin: 0x%lx\n", selwin);
+		}
 	}
 
 	if ((xrandr || xrandr_maybe) && !did_xrandr) {
@@ -158,8 +394,16 @@ static void initialize_xevents(int reset) {
 		did_clipboard_atom = 1;
 	}
 	if (xfixes_present && use_xfixes && !did_xfixes) {
-		initialize_xfixes();
-		did_xfixes = 1;
+		/*
+		 * We try to delay creating initializing xfixes until
+		 * we are past the display manager, due to Xorg bug:
+		 * http://bugs.freedesktop.org/show_bug.cgi?id=18451
+		 */
+		if (guess_dm_gone(5, 45)) {
+			initialize_xfixes();
+			did_xfixes = 1;
+			if (! quiet) rfbLog("called initialize_xfixes()\n");
+		}
 	}
 	if (xdamage_present && !did_xdamage) {
 		initialize_xdamage();
@@ -180,7 +424,7 @@ static void print_xevent_bases(void) {
 	fprintf(stderr, "  SelClear=%d, Expose=%d\n", SelectionClear, Expose);
 }
 
-void get_prop(char *str, int len, Atom prop) {
+void get_prop(char *str, int len, Atom prop, Window w) {
 	int i;
 #if !NO_X11
 	Atom type;
@@ -203,9 +447,12 @@ void get_prop(char *str, int len, Atom prop) {
 #else
 
 	slen = 0;
-	
+	if (w == None) {
+		w = DefaultRootWindow(dpy);
+	}
+
 	do {
-		if (XGetWindowProperty(dpy, DefaultRootWindow(dpy),
+		if (XGetWindowProperty(dpy, w,
 		    prop, nitems/4, len/16, False,
 		    AnyPropertyType, &type, &format, &nitems, &bytes_after,
 		    &data) == Success) {
@@ -497,6 +744,7 @@ static void grab_buster_watch(int parent, char *dstr) {
 	char propval[200];
 	int ev, er, maj, min;
 	int db = 0;
+	char *ticker_str = "X11VNC_TICKER";
 
 	RAWFB_RET_VOID
 
@@ -534,7 +782,10 @@ static void grab_buster_watch(int parent, char *dstr) {
 
 	usleep(3 * sleep);
 
-	ticker_atom = XInternAtom(dpy, "X11VNC_TICKER", False);
+	if (getenv("X11VNC_TICKER")) {
+		ticker_str = getenv("X11VNC_TICKER");
+	}
+	ticker_atom = XInternAtom(dpy, ticker_str, False);
 	if (! ticker_atom) {
 		fprintf(stderr, "grab_buster_watch: no ticker atom\n");
 		return;
@@ -552,7 +803,7 @@ static void grab_buster_watch(int parent, char *dstr) {
 			break;
 		}
 
-		get_prop(propval, 128, ticker_atom);
+		get_prop(propval, 128, ticker_atom, None);
 		if (db) fprintf(stderr, "got_prop:   %s\n", propval);
 
 		if (!process_watch(propval, parent, db)) {
@@ -628,7 +879,11 @@ void sync_tod_with_servertime(void) {
 	}
 
 	if (! ticker_atom) {
-		ticker_atom = XInternAtom(dpy, "X11VNC_TICKER", False);
+		char *ticker_str = "X11VNC_TICKER";
+		if (getenv("X11VNC_TICKER")) {
+			ticker_str = getenv("X11VNC_TICKER");
+		}
+		ticker_atom = XInternAtom(dpy, ticker_str, False);
 	}
 	if (! ticker_atom) {
 		return;
@@ -691,7 +946,9 @@ void check_keycode_state(void) {
 	 * in sync with the Xserver at every instant of time.
 	 */
 	if (now > last_check + delay && now > last_keyboard_input + noinput) {
+		X_LOCK;
 		init_track_keycode_state();
+		X_UNLOCK;
 		last_check = now;
 	}
 }
@@ -788,9 +1045,9 @@ void check_local_grab(void) {
 
 void check_autorepeat(void) {
 	static time_t last_check = 0;
+	static int idle_timeout = -300, idle_reset = 0;
 	time_t now = time(NULL);
-	int autorepeat_is_on, autorepeat_initially_on, idle_timeout = 300;
-	static int idle_reset = 0;
+	int autorepeat_is_on, autorepeat_initially_on;
 
 	if (! no_autorepeat || ! client_count) {
 		return;
@@ -801,6 +1058,15 @@ void check_autorepeat(void) {
 
 	if (unixpw_in_progress) return;
 
+	if (idle_timeout < 0) {
+		if (getenv("X11VNC_IDLE_TIMEOUT")) {
+			idle_timeout = atoi(getenv("X11VNC_IDLE_TIMEOUT"));
+		}
+		if (idle_timeout < 0) {
+			idle_timeout = -idle_timeout;
+		}
+	}
+
 	last_check = now;
 
 	autorepeat_is_on = get_autorepeat_state();
@@ -818,7 +1084,7 @@ void check_autorepeat(void) {
 		if (! autorepeat_is_on && autorepeat_initially_on) {
 			static time_t last_msg = 0;
 			static int cnt = 0;
-			if (now > last_msg + idle_timeout && cnt++ < 5) {
+			if (now > last_msg + idle_timeout && cnt++ < 10) {
 				rfbLog("idle keyboard:   turning X autorepeat"
 				    " back on.\n");
 				last_msg = now;
@@ -827,10 +1093,55 @@ void check_autorepeat(void) {
 			idle_reset = 1;
 		}
 	} else {
+		if (idle_reset) {
+			int i, state[256], didmsg = 0, pressed = 0;
+			int mwt = 600, mmax = 20;
+			static int msgcnt = 0;
+			static double lastmsg = 0.0;
+
+			for (i=0; i<256; i++) {
+				state[i] = 0;
+			}
+			if (use_threads) {X_LOCK;}
+			get_keystate(state);
+			if (use_threads) {X_UNLOCK;}
+
+			for (i=0; i<256; i++) {
+				if (state[i] != 0) {
+					/* better wait until all keys are up  */
+					pressed++;
+					if (msgcnt < mmax || dnow() > lastmsg + mwt) {
+						char *str = "unset";
+#if !NO_X11
+						if (use_threads) {X_LOCK;}
+						str = XKeysymToString(XKeycodeToKeysym(dpy, i, 0));
+						if (use_threads) {X_UNLOCK;}
+#endif
+						str = str ? str : "nosymbol";
+						didmsg++;
+						rfbLog("active keyboard: waiting until "
+						    "all keys are up. key_down=%d %s.  "
+						    "If the key is inaccessible via keyboard, "
+						    "consider 'x11vnc -R clear_all'\n", i, str);
+					}
+				}
+			}
+			if (didmsg > 0) {
+				msgcnt++;
+				if (msgcnt == mmax) {
+					rfbLog("active keyboard: last such "
+					    "message for %d secs.\n", mwt);
+				}
+				lastmsg = dnow();
+			}
+			if (pressed > 0) {
+				return;
+			}
+		}
 		if (idle_reset) {
 			static time_t last_msg = 0;
 			static int cnt = 0;
-			if (now > last_msg + idle_timeout && cnt++ < 5) {
+			if (now > last_msg + idle_timeout && cnt++ < 10) {
 				rfbLog("active keyboard: turning X autorepeat"
 				    " off.\n");
 				last_msg = now;
@@ -995,38 +1306,99 @@ void check_xevents(int reset) {
 		last_call = now;
 	}
 
-	/* check for CUT_BUFFER0 and VNC_CONNECT changes: */
-	if (XCheckTypedEvent(dpy, PropertyNotify, &xev)) {
-		if (xev.type == PropertyNotify) {
-			if (xev.xproperty.atom == XA_CUT_BUFFER0) {
-				/*
-				 * Go retrieve CUT_BUFFER0 and send it.
-				 *
-				 * set_cutbuffer is a flag to try to avoid
-				 * processing our own cutbuffer changes.
-				 */
-				if (have_clients && watch_selection
-				    && ! set_cutbuffer) {
-					cutbuffer_send();
-					sent_some_sel = 1;
+	if (freeze_when_obscured) {
+		if (XCheckTypedEvent(dpy, VisibilityNotify, &xev)) {
+			if (xev.type == VisibilityNotify && xev.xany.window == subwin) {
+				int prev = subwin_obscured;
+				if (xev.xvisibility.state == VisibilityUnobscured) {
+					subwin_obscured = 0;
+				} else if (xev.xvisibility.state == VisibilityPartiallyObscured) {
+					subwin_obscured = 1;
+				} else {
+					subwin_obscured = 2;
 				}
-				set_cutbuffer = 0;
-			} else if (vnc_connect && vnc_connect_prop != None
-		    	    && xev.xproperty.atom == vnc_connect_prop) {
-				/*
-				 * Go retrieve VNC_CONNECT string.
-				 */
-				read_vnc_connect_prop(0);
-			} else if (vnc_connect && x11vnc_remote_prop != None
-		    	    && xev.xproperty.atom == x11vnc_remote_prop) {
-				/*
-				 * Go retrieve X11VNC_REMOTE string.
-				 */
-				read_x11vnc_remote_prop(0);
-
-
+				rfbLog("subwin_obscured: %d -> %d\n", prev, subwin_obscured);
 			}
-			set_prop_atom(xev.xproperty.atom);
+		}
+	}
+
+	/* check for CUT_BUFFER0, VNC_CONNECT, X11VNC_REMOTE changes: */
+	if (XCheckTypedEvent(dpy, PropertyNotify, &xev)) {
+		int got_cutbuffer = 0;
+		int got_vnc_connect = 0;
+		int got_x11vnc_remote = 0;
+		static int prop_dbg = -1;
+
+		/* to avoid piling up between calls, read all PropertyNotify now */
+		do {
+			if (xev.type == PropertyNotify) {
+				if (xev.xproperty.atom == XA_CUT_BUFFER0) {
+					got_cutbuffer++;
+				} else if (vnc_connect && vnc_connect_prop != None
+				    && xev.xproperty.atom == vnc_connect_prop) {
+					got_vnc_connect++;
+				} else if (vnc_connect && x11vnc_remote_prop != None
+				    && xev.xproperty.atom == x11vnc_remote_prop) {
+					got_x11vnc_remote++;
+				}
+				set_prop_atom(xev.xproperty.atom);
+			}
+		} while (XCheckTypedEvent(dpy, PropertyNotify, &xev));
+
+		if (prop_dbg < 0) {
+			prop_dbg = 0;
+			if (getenv("PROP_DBG")) {
+				prop_dbg = 1;
+			}
+		}
+
+		if (prop_dbg && (got_cutbuffer > 1 || got_vnc_connect > 1 || got_x11vnc_remote > 1)) {
+			static double lastmsg = 0.0;
+			static int count = 0;
+			double now = dnow();
+
+			if (1 && now > lastmsg + 300.0) {
+				if (got_cutbuffer > 1) {
+					rfbLog("check_xevents: warning: %d cutbuffer events since last check.\n", got_cutbuffer);
+				}
+				if (got_vnc_connect > 1) {
+					rfbLog("check_xevents: warning: %d vnc_connect events since last check.\n", got_vnc_connect);
+				}
+				if (got_x11vnc_remote > 1) {
+					rfbLog("check_xevents: warning: %d x11vnc_remote events since last check.\n", got_x11vnc_remote);
+				}
+				count++;
+				if (count >= 3) {
+					lastmsg = now;
+					count = 0;
+				}
+			}
+		}
+
+		if (got_cutbuffer)  {
+			/*
+			 * Go retrieve CUT_BUFFER0 and send it.
+			 *
+			 * set_cutbuffer is a flag to try to avoid
+			 * processing our own cutbuffer changes.
+			 */
+			if (have_clients && watch_selection && !set_cutbuffer) {
+				cutbuffer_send();
+				sent_some_sel = 1;
+			}
+			set_cutbuffer = 0;
+		} 
+		if (got_vnc_connect) {
+			/*
+			 * Go retrieve VNC_CONNECT string.
+			 */
+			read_vnc_connect_prop(0);
+		} 
+		if (got_x11vnc_remote) {
+			/*
+			 * Go retrieve X11VNC_REMOTE string.
+			 */
+			read_x11vnc_remote_prop(0);
 		}
 	}
 
@@ -1046,7 +1418,7 @@ void check_xevents(int reset) {
 	}
 #endif
 #if LIBVNCSERVER_HAVE_LIBXFIXES
-	if (xfixes_present && use_xfixes && xfixes_base_event_type) {
+	if (xfixes_present && use_xfixes && xfixes_first_initialized && xfixes_base_event_type) {
 		if (XCheckTypedEvent(dpy, xfixes_base_event_type +
 		    XFixesCursorNotify, &xev)) {
 			got_xfixes_cursor_notify++;
@@ -1124,7 +1496,7 @@ void check_xevents(int reset) {
 			req = "CLIPBOARD";
 		}
 		if (which != 0 && ! own && have_clients &&
-		    XGetSelectionOwner(dpy, atom) != None) {
+		    XGetSelectionOwner(dpy, atom) != None && selwin != None) {
 			XConvertSelection(dpy, atom, XA_STRING, XA_STRING,
 			    selwin, CurrentTime);
 			if (debug_sel) {
@@ -1288,6 +1660,10 @@ extern int rawfb_vnc_reflect;
 void xcut_receive(char *text, int len, rfbClientPtr cl) {
 	allowed_input_t input;
 
+	if (threads_drop_input) {
+		return;
+	}
+
 	if (unixpw_in_progress) {
 		rfbLog("xcut_receive: unixpw_in_progress, skipping.\n");
 		return;
@@ -1306,33 +1682,91 @@ void xcut_receive(char *text, int len, rfbClientPtr cl) {
 	if (!input.clipboard) {
 		return;
 	}
+	INPUT_LOCK;
+
+	if (remote_prefix != NULL && strstr(text, remote_prefix) == text) {
+		char *result, *rcmd = text + strlen(remote_prefix);
+		char *tmp = (char *) calloc(len + 8, 1);
+
+		if (strstr(rcmd, "cmd=") != rcmd && strstr(rcmd, "qry=") != rcmd) {
+			strcat(tmp, "qry=");
+		}
+		strncat(tmp, rcmd, len - strlen(remote_prefix));
+		rfbLog("remote_prefix command: '%s'\n", tmp);
+
+		if (use_threads) {
+			if (client_connect_file) {
+				FILE *f = fopen(client_connect_file, "w");
+				if (f) {
+					fprintf(f, "%s\n", tmp);
+					fclose(f);
+					free(tmp);
+					INPUT_UNLOCK;
+					return;
+				}
+			}
+			if (vnc_connect) {
+				sprintf(x11vnc_remote_str, "%s", tmp);
+				free(tmp);
+				INPUT_UNLOCK;
+				return;
+			}
+		}
+		INPUT_UNLOCK;
+
+
+		result = process_remote_cmd(tmp, 1);
+		if (result == NULL ) {
+			result = strdup("null");
+		} else if (!strcmp(result, "")) {
+			free(result);
+			result = strdup("none");
+		}
+		rfbLog("remote_prefix result:  '%s'\n", result);
+
+		free(tmp);
+		tmp = (char *) calloc(strlen(remote_prefix) + strlen(result) + 1, 1);
+
+		strcat(tmp, remote_prefix);
+		strcat(tmp, result);
+		free(result);
+
+		rfbSendServerCutText(screen, tmp, strlen(tmp));
+		free(tmp);
+
+		return;
+	}
 
 	if (! check_sel_direction("recv", "xcut_receive", text, len)) {
+		INPUT_UNLOCK;
 		return;
 	}
 
 #ifdef MACOSX
 	if (macosx_console) {
 		macosx_set_sel(text, len);
+		INPUT_UNLOCK;
 		return;
 	}
 #endif
 
 	if (rawfb_vnc_reflect) {
 		vnc_reflect_send_cuttext(text, len);
+		INPUT_UNLOCK;
 		return;
 	}
 
 	RAWFB_RET_VOID
 
 #if NO_X11
+	INPUT_UNLOCK;
 	return;
 #else
 
 	X_LOCK;
 
 	/* associate this text with PRIMARY (and SECONDARY...) */
-	if (set_primary && ! own_primary) {
+	if (set_primary && ! own_primary && selwin != None) {
 		own_primary = 1;
 		/* we need to grab the PRIMARY selection */
 		XSetSelectionOwner(dpy, XA_PRIMARY, selwin, CurrentTime);
@@ -1342,7 +1776,7 @@ void xcut_receive(char *text, int len, rfbClientPtr cl) {
 		}
 	}
 
-	if (set_clipboard && ! own_clipboard && clipboard_atom != None) {
+	if (set_clipboard && ! own_clipboard && clipboard_atom != None && selwin != None) {
 		own_clipboard = 1;
 		/* we need to grab the CLIPBOARD selection */
 		XSetSelectionOwner(dpy, clipboard_atom, selwin, CurrentTime);
@@ -1384,6 +1818,7 @@ void xcut_receive(char *text, int len, rfbClientPtr cl) {
 	XFlush_wr(dpy);
 
 	X_UNLOCK;
+	INPUT_UNLOCK;
 
 	set_cutbuffer = 1;
 #endif	/* NO_X11 */
@@ -1403,8 +1838,16 @@ void kbd_release_all_keys(rfbClientPtr cl) {
 #if NO_X11
 	return;
 #else
+	if (use_threads) {
+		X_LOCK;
+	}
+
 	clear_keys();
 	clear_modifiers(0);
+
+	if (use_threads) {
+		X_UNLOCK;
+	}
 #endif
 }
 
@@ -1581,10 +2024,11 @@ static void try_local_chat_window(void) {
 			return;
 		}
 
+		/* mutex */
 		new_save = screen->newClientHook;
 		screen->newClientHook = new_client_chat_helper;
 
-		chat_window_client = rfbNewClient(screen, sock);
+		chat_window_client = create_new_client(sock, 1);
 
 		screen->newClientHook = new_save;
 
@@ -1646,6 +2090,9 @@ void set_text_chat(rfbClientPtr cl, int len, char *txt) {
 		if (cl2->state != RFB_NORMAL) {
 			continue;
 		}
+
+		SEND_LOCK(cl2);
+
 		if (ulen == rfbTextChatOpen) {
 			rfbSendTextChatMessage(cl2, rfbTextChatOpen, "");
 		} else if (ulen == rfbTextChatClose) {
@@ -1657,12 +2104,16 @@ void set_text_chat(rfbClientPtr cl, int len, char *txt) {
 		} else if (len <= rfbTextMaxSize) {
 			rfbSendTextChatMessage(cl2, len, txt);
 		}
+
+		SEND_UNLOCK(cl2);
 	}
 	rfbReleaseClientIterator(iter);
 
 	if (ulen == rfbTextChatClose && cl != NULL) {
 		/* not clear what is going on WRT close and finished... */
+		SEND_LOCK(cl);
 		rfbSendTextChatMessage(cl, rfbTextChatFinished, "");
+		SEND_UNLOCK(cl);
 	}
 #endif
 }
